/************************************************************************
 *  Revision 1.0 -- converted subroutines of video board test program to
 *      a rudamentary demo program and sample code.
 *  Revision 1.1 -- the forgotten tables for 80x44 monochrome and 80x50
 *      color text modes added to the end (not best place, but people
 *      are already aware of the first 22 modes). [17 Oct 86]
 *  Revision 1.2 -- the table entries marked (*) in the configuration
 *	tables have been moved about to match the modes used by Tseng
 *	Labs in their driver BIGSCR.  The STB driver INT10 has also
 *	been modified to reflect this new standard. [3 Feb 87]
 *  Revision 2.0 -- modified initialization code and tables for use
 *	with the EGA MultiRes II, demonstrating its modes. [22 April 87]
 *  Revision 2.1 -- added conditional inclusion of the VGA 256 color mode
 *	in exchange for the 132 column modes [July 9, 1987]
 *  Revision 2.2 -- added "automatic" adaptation of the default demo
 *	sequence so only valid screens are used with each monitor type
 *	when invoked without a parameter list [August 19, 1987]
 *  Revision 2.3 -- removed "holes", etc, in the 40 column display (left
 *	them in the 132 column display where they are not so obnoxious
 *	and thus we can still show the full character set!)
 *	[August 20, 1987]
 *  Revision 2.4 -- added compile flag to delete all 132 column and high
 *	resolution graphics displays for the STB EGA [September 17, 1987]
 *  Revision 2.5 -- corrected rainbow display for VGA and added support for
 *	EGA/VGA text with attributes (including transparent background)
 *	[October 7, 1987]
 *  Revision 2.6 -- "prettified" the transparent text display to make it
 *	more consistent between the various color displays and fixed a bug
 *	in the 2-color graphics text output code. Added XOR text output
 *	code [October 9, 1987]
 *  Revision 2.7 -- added support for the "variable" resolution mode (11)
 *	[November 9, 1987]
 *  Revision 2.8 -- cleaned up background/foreground colors for the 80 column
 *	text modes and the default sequences for the new VGA products
 *	[December 18, 1987]
 *  Revision 2.9 -- made some minor corrections to the graphics text routines
 *	including support for fonts other than 8x8, and added code to handle
 *	then VGA/em 1024x768 graphics mode (as mode 11) [December 30, 1987]
 *  Revision 2.10 -- added error traps in fill code so partially populated
 *	boards will not cause the program to hang in an infinite loop and
 *	added a "Usage:" message if an illegal option (/? for example) is
 *	entered.  Also, added a flag (/D) to shorten 256 color displays when
 *	used with digital monitors. [January 20, 1988]
 *  Revision 2.11 -- reduced the number of default screens (so dumb salesmen
 *	won't get bad screens when showing off cards or monitors!) and
 *	made video mode 12 self-configuring like mode 11 [February 25, 1988]
 *  Revision 2.12 -- added hardware test to keep 256 color screens off the
 *	demo sequence if no analog palatte chip is present, and cleaned up
 *	the code to skip cycling colors if the "/D" flag was entered
 *	[March 4, 1988]
 *  Revision 2.13 -- added "undef"s to demo.c to permit compilation with
 *	Turbo C [March 6, 1988]
 *  Revision 2.14 -- added video mode 24h (132x28) for DCA, et al.
 *	[April 12, 1988]
 *  Revision 2.15 -- modified "default_ega" setup to generate only legal
 *	screens for the Autoega [July 21, 1988]
 *  Revision 2.16 -- made the STB on the color bar optional (Packard Bell)
 *	-- this also eliminates 512K modes except for 1024x768[October 2, 1988]
 *  Revision 2.17 -- made the 8514 compatible modes selectable with option
 *	1024 rather than VGA8514 and added memory test for 512K modes
 *	[October 18, 1988]
 *  Revision 2.18 -- added support for the Cirrus GD 510A/520A chip set
 *	"standard" modes 64-66, 98-100 and 112-114, as well as the IBM/Cirrus
 *	high resolution 256-color memory mapping [November 8, 1988]
 *  Revision 2.19 -- added Micron Technology as alternate to STB in text
 *	strings for graphics pages [November 10, 1988]
 *  Revision 2.20 -- added new text modes for non-standard line counts
 *	[November 23, 1988]
 *  Revision 2.21 -- moved customization code out of main program to
 *	minimize recompile time [January 23, 1989]
 *  Revision 3.00 -- redid the whole mess, so we can use it as a "dumb"
 *	demo [May 18, 1989]
 *  Revision 3.01 -- added support for the new ET4000 memory mapping in
 *	high resolution 16 and 256 color modes [January 21, 1990]
 *  Revision 3.02 -- corrected ET4000 code for 1024x768 [March 14, 1990]
 ***********************************************************************/
#include <stdio.h>
#include <dos.h>
#undef peek
#undef pokeb
#undef pokew

#define EGA      0
#define VGA      1
#define CHIPS435 2
#define CHIPS441 3
#define CIRRUS   4
#define CHIPS451 5
#define TSENG    6
#define ET4000   7

char *ctype[] = {"EGA", "Standard VGA", "C&T 435", "C&T 441", "Cirrus 510/520",
      "C&T 451/452", "Tseng Labs ET3000", "Tseng Labs ET4000"};

/***********************************************************************
 *  Function call explanations:
 *
 *  (*color_c)(row, col, attribute, character)
 *	writes the character "character" on text row "row", character
 *	column "col" using the color specified in the low byte of
 *	"attribute".  If bit 8 of "attribute" is a 1, the background
 *	color is undisturbed, otherwise it is set to color 0 (Black).
 *	If bit 9 of "attribute" is 1, the character is exclusived-ORed
 *	with current contents of the character cell rather than normal
 *	"replace" option.  If both bits 8 and 9 are set to 1, the operation
 *	is UNDEFINED and in text modes bits 8 and 9 are ignored.
 *
 ***********************************************************************/

#ifdef WATCOM
volatile void (*dot)(volatile int, volatile int); /* the current pixel output routine */
volatile unsigned int (*get_dot)(); /* the current read  pixel routine */
volatile int (*find_color)();	    /* the current find pixel of color routine */

int cdecl ega_bit(), cdecl ega_get(), cdecl egacolor();
int cdecl et4000_bit(), cdecl et4000_get(), cdecl et4000color();
int cdecl one_bit(), cdecl one_get(), cdecl onecolor();
int cdecl two_bit(), cdecl two_get(), cdecl twocolor();
int cdecl four_bit(), cdecl four_get(), cdecl fourcolor();
int cdecl ibm_bit(), cdecl ibm_get(), cdecl ibmcolor();
void cdecl filler(), cdecl pokeb(), cdecl pokew();
int cdecl peek();
void cdecl set_palette();
volatile void cdecl vidmode();
volatile void cdecl wrt_ega(volatile int, int);
unsigned char far * cdecl getfont();
#else
void (*dot)();			/* the current pixel output routine */
unsigned int (*get_dot)();	/* the current read  pixel routine */
int (*find_color)();		/* the current find pixel of color routine */

int ega_bit(), ega_get(), egacolor();
int et4000_bit(), et4000_get(), et4000color();
int one_bit(), one_get(), onecolor();
int two_bit(), two_get(), twocolor();
int four_bit(), four_get(), fourcolor();
int ibm_bit(), ibm_get(), ibmcolor();
void filler(), pokeb(), pokew();
int peek();
void set_palette();
volatile void vidmode();
volatile void wrt_ega();
unsigned char far *getfont();
#endif

unsigned int rr();

char version[] = {
      "EGA/VGA DEMO, Version 2.19, (c) Copyright 1988, STB Systems, Inc."
      };

int chiptype;
unsigned int regen;
unsigned int memend;
unsigned int crtc;
unsigned int mode_reg;
unsigned int color_reg;
unsigned int limit1, limit2;
unsigned int row_start[1024];	/* start address of each scan row */
unsigned char bank_sel[1024];	/* page used for each scan row */
unsigned int expand_left[256];	/* graphics text char pattern expand tables */
unsigned int expand_right[256];	/*   that expand 1 bit per pixel into 2 or 4 bpp */
unsigned char far *fontptr;	/* character generator table base address */
unsigned int text_height = 8;   /* graphics char height */
int (*color_c)();		/* the current write char */
unsigned int clear_word;	/* word value to clear the buffer with */
unsigned int aspect;

int extmode[20]; /* list of video mode numbers to test */

unsigned int xmax;
unsigned int ymax;
unsigned int zmax;
unsigned int bits_per_pixel;
unsigned int pixels_per_byte;
unsigned int textcols;
unsigned int textrows;
unsigned int max_scan;
int vmode;
int bflag, lflag;
int digital_flag;
int is8514;
int is512k;

int color;			/* current drawing color */
unsigned int *color_table;	/* pointer to current color expand table */
unsigned char *get_shift;	/* pointer to read pixel shift count table */

/* tables for one bit per pixel */
unsigned char ega_mask[] =   { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };
unsigned int  ega_color[] =  { /* EGA 16-color modes expand a color to itself */
   0x0000, 0x0101, 0x0202, 0x0303, 0x0404,
   0x0505, 0x0606, 0x0707, 0x0808, 0x0909,
   0x0A0A, 0x0B0B, 0x0C0C, 0x0D0D, 0x0E0E, 0x0F0F};
unsigned char one_mask[] =   { 0x7f, 0xbf, 0xdf, 0xef, 0xf7, 0xfb, 0xfd, 0xfe };
unsigned int  one_color[] =  { /* color pattern expand table */
   0x0000, 0xFFFF, 0xFFFF, 0xFFFF,
   0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF,
   0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF,
   0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF
};
char one_shift[] =  { 7, 6, 5, 4, 3, 2, 1, 0 };
int ibm_map[] = {0,0x0802,0x0402,0,0x0202,0,0,0,0x0102,0,0,0,0,0,0,0};

/* tables for two bits per pixel */
unsigned char two_mask[] =   { 0x3F, 0xCF, 0xF3, 0xFC };
unsigned int  two_color[] =  {
   0x0000, 0x5555, 0xAAAA, 0xFFFF,
   0xFFFF, 0x5555, 0xAAAA, 0xFFFF,
   0xFFFF, 0x5555, 0xAAAA, 0xFFFF,
   0xFFFF, 0x5555, 0xAAAA, 0xFFFF
};
char two_shift[] =  { 6, 4, 2, 0 };

/* tables for four bits per pixel */
unsigned char four_mask[] =   { 0x0F, 0xF0 };
unsigned int  four_color[] =  {
   0x0000, 0x1111, 0x2222, 0x3333,
   0x4444, 0x5555, 0x6666, 0x7777,
   0x8888, 0x9999, 0xAAAA, 0xBBBB,
   0xCCCC, 0xDDDD, 0xEEEE, 0xFFFF
};
char four_shift[] =  { 4, 0 };

int hours, minutes, seconds, msecs, month, day, year, weekday;

/* Stuff for drawing test patterns */
unsigned char bar_pattern[] = {	/* blocks for drawing a color bar */
   0xDB, 0xDB, 0xDB, 0xDB, 0xDB, 0
};

char *colors[] = {
   "BLACK",  "BLUE",     "GREEN",   "CYAN",
   "RED",    "MAGENTA",  "BROWN",  "WHITE",
   "GREY",   "BLUE",     "GREEN",   "CYAN",
   "RED",    "MAGENTA",  "YELLOW",  "WHITE"
};

char *four_colors[4][4] = {  /* four color mode pallettes */
   {"BKG",  "GREEN",    "RED",       "BROWN" }, /* main pallettes */
   {"BKG",  "CYAN",     "MAGENTA",   "WHITE" },

   {"BKG",  "GREEN",    "RED",       "BROWN" }, /* alternate pallettes */
   {"BKG",  "CYAN",     "RED",       "WHITE" }
};

char *color_abbrev[] = {
   "BLK",   "BLU",   "GRN",   "CYN",
   "RED",   "MAG",   "BRN",   "WHT",
   "GRY",   "BLU",   "GRN",   "CYN",
   "RED",   "MAG",   "YEL",   "WHT"
};

struct mode_table {
   int pixels;
   int lines;
   unsigned char chars;
   unsigned char rows;
   int colors;
   unsigned int segmt;
   unsigned char charsize;
   unsigned char text;
   unsigned int size;
   unsigned char frames;
   unsigned char bits;
   unsigned int addr;
   unsigned char blink;
   };

struct mode_table *setup;
extern struct mode_table mode_tbl[];

char pal_tab[257*3];		/* Maximum size */

int rainbow[256];		/* Use to generate the rainbow palette */

main(argc, argv)
   int argc;
   char *argv[];
   {
   int i, j, x;
   int monostate;

   monostate = 0;
   if (peek(0x463,0) == 0x3B4)
      monostate = 1;
   chiptype = chipset();
   if (argc == 2 && stricmp(argv[1], "test") == 0) {
      printf("Video chip is a %s\n", ctype[chiptype]);
      exit(99);
      }
   is512k = 0;
   is8514 = 0;
   if (chiptype > CIRRUS) {
      if ((inp(0x3C2) & 0x60) == 0)
         is8514 = 1;
      if (mem512())
         is512k = 1;
      }
   /*
    * Configure mode 11
    */
   i = peek(0x4A8,0);
   j = peek(0x4AA,0);
   x = peek(i,j);
   j = peek(i+2,j);

   i = x + 11*64;

   mode_tbl[11].chars = peek(i,j) & 0xFF;
   mode_tbl[11].rows = (peek(i+1,j) & 0xFF) + 1;
   mode_tbl[11].charsize = peek(i+2,j) & 0xFF;
   mode_tbl[11].pixels = mode_tbl[11].chars * 8;
   mode_tbl[11].lines = mode_tbl[11].rows * mode_tbl[11].charsize;
   mode_tbl[11].size = peek(i+3,j);
   x = peek(i+4,j) & 0xFF;
   if (x > 3)
      mode_tbl[11].frames = 0x100/x;
   else
      mode_tbl[11].frames = 0;
   mode_tbl[11].text = peek(i+8,j) & 0x01;
   mode_tbl[11].addr = 0x3B4 + (peek(i+9,j) & 0x01)*0x20;
   mode_tbl[11].blink = (peek(i+51,j) & 0x08) >> 3;
   mode_tbl[11].bits = 1;
   mode_tbl[11].colors = 16;

   /*
    * Configure mode 12
    */
   i += 64;
   mode_tbl[12].chars = peek(i,j) & 0xFF;
   mode_tbl[12].rows = (peek(i+1,j) & 0xFF) + 1;
   mode_tbl[12].charsize = peek(i+2,j) & 0xFF;
   mode_tbl[12].pixels = mode_tbl[12].chars * 8;
   mode_tbl[12].lines = mode_tbl[12].rows * mode_tbl[12].charsize;
   mode_tbl[12].size = peek(i+3,j);
   x = peek(i+4,j) & 0xFF;
   if (x > 3)
      mode_tbl[12].frames = 0x100/x;
   else
      mode_tbl[12].frames = 0;
   mode_tbl[12].text = peek(i+8,j) & 0x01;
   mode_tbl[12].addr = 0x3B4 + (peek(i+9,j) & 0x01)*0x20;
   mode_tbl[12].blink = (peek(i+51,j) & 0x08) >> 3;
   mode_tbl[12].bits = 1;
   mode_tbl[12].colors = 16;

   for (i=0; i < 256; i++)
      rainbow[i] = 0;
   sine_tab(129, 63, rainbow);
   i = peek(0x488,0) & 0x0F;
   j = 0;
   bflag = 0;
   lflag = 0;
   digital_flag = 1;
   if (color256())
      digital_flag = 0;
   while (argv[1][0] == '/' || argv[1][0] == '-') {
      if (argv[1][1] == 'L' || argv[1][1] == 'l') {
         lflag = 1;
         argv++;
         argc--;
         }
      else if (argv[1][1] == 'B' || argv[1][1] == 'b') {
         if ((bflag = atoi(argv[1]+2)) == 0)
            bflag = 14;
         argv++;
         argc--;
         }
      else if (argv[1][1] == 'D' || argv[1][1] == 'd') {
         digital_flag ^= 1;
         argv++;
         argc--;
         }
      else {
usage:
         printf("Usage: DEMO [/B[nn]] [/L] [/D] {VGA|EGA|RGB|MONO|1024}\n");
         printf("   or  DEMO [/B[nn]] [/L] [/D] list-of-valid-modes\n");
         printf("   or  DEMO [/B[nn]] [/L] [/D]\n");
#ifdef CITIZEN
         printf("   where /B3 sets the overscan area (if any) to colour 3\n");
         printf("         /L sets the demo program into \"looping\" mode\n");
         printf("         /D converts 256 colour analog modes to 64 colour digital modes\n");
#else
         printf("   where /B3 sets the overscan area (if any) to color 3\n");
         printf("         /L sets the demo program into \"looping\" mode\n");
         printf("         /D converts 256 color analog modes to 64 color digital modes\n");
#endif
         printf("   and if a list of modes is not given, the default monitor type\n");
         printf("   can be overridden by explicitly identifying a monitor on the\n");
         printf("   command line.\n");
         exit(0);
         }
      }
   if (chiptype == CIRRUS) {
      if (argc <= 1) {
         if (monostate)
            goto default_mono;
         extmode[0] = 0;	/* 40 col x 25 line color text */
         extmode[1] = 3;	/* 80 col x 25 line color text */
         extmode[2] = 4;	/* 320 x 200 4-color graphics */
         extmode[3] = 6;	/* 640 x 200 2-color graphics */
         extmode[4] = 13;	/* 320 x 200 16-color graphics */
         extmode[5] = 14;	/* 640 x 200 16-color graphics */
         extmode[6] = 16;	/* 640 x 350 16-color graphics */
         extmode[7] = 17;	/* 640 x 480 2-color graphics */
         extmode[8] = 18;	/* 640 x 480 16-color graphics */
         extmode[9] = 19;	/* 320 x 200 256-color graphics */
         argc = 11;
         goto do_demo;
         }
      goto single;
      }
   if (argc <= 1) {
      if (monostate)
         goto default_mono;
      if (i != 3 && i != 9)
         goto default_color;
      if ((peek(0x489,0) & 0x10) == 0)
         goto default_ega;
      goto default_vga;
      }
   if(argc == 2) {
#ifdef CITIZEN
      if (strcmp(argv[1], "colour") == 0 || strcmp(argv[1], "COLOUR") == 0 ||
#else
      if (strcmp(argv[1], "color") == 0 || strcmp(argv[1], "COLOR") == 0 ||
#endif
            strcmp(argv[1], "cga") == 0 || strcmp(argv[1], "CGA") == 0 ||
            strcmp(argv[1], "rgb") == 0 || strcmp(argv[1], "RGB") == 0) {
default_color:
         extmode[0] = 0;	/* 40 col x 25 line color text */
         extmode[1] = 3;	/* 80 col x 25 line color text */
         extmode[2] = 4;	/* 320 x 200 4-color graphics */
         extmode[3] = 6;	/* 640 x 200 2-color graphics */
         extmode[4] = 13;	/* 320 x 200 16-color graphics */
         extmode[5] = 14;	/* 640 x 200 16-color graphics */
         argc = 7;
         }
      else if (strcmp(argv[1], "mono") == 0 || strcmp(argv[1], "MONO") == 0) {
default_mono:
         extmode[0] = 7;	/* 80 col x 25 line monochrome text */
         extmode[1] = 15;	/* 640x350 (EGA) Monochrome graphics */
         argc = 3;
         }
#ifdef CITIZEN
      else if (stricmp(argv[1], "vga") == 0 || stricmp(argv[1], "msync") == 0 ||
            strcmp(argv[1], "1024") == 0) {
default_vga:
         extmode[0] = 0;	/* 40 col x 25 line color text */
         extmode[1] = 3;	/* 80 col x 25 line color text */
         extmode[2] = 34;	/* 132 col x 44 line color text */
         extmode[3] = 13;	/* 320 x 200 16-color graphics */
         extmode[4] = 19;	/* 320 x 200 256-color graphics */
         extmode[5] = 18;	/* 640 x 480 16-color graphics */
         extmode[6] = 46;	/* 640 x 480 256-color graphics */
         extmode[7] = 41;	/* 800 x 600 16-color graphics */
         extmode[8] = 48;	/* 800 x 600 256-color graphics */
         extmode[9] = 55;	/* 1024 x 768 16-color graphics */
         argc = 11;
         if (chiptype == ET4000) {
            extmode[10] = 56;	/* 1024 x 768 16-color graphics */
            argc = 12;
            }
         if (chiptype < CHIPS451) {
            printf("Cannot display 1024x768 resolution on this video adapter\n");
            goto usage;
            }
         if (is512k == 0) {
            printf("Cannot display 1024x768 resolution without 512K bytes of video memory\n");
            goto usage;
            }
         }
#else
      else if (strcmp(argv[1], "vga") == 0 || strcmp(argv[1], "VGA") == 0) {
default_vga:
         extmode[0] = 0;	/* 40 col x 25 line color text */
         extmode[1] = 3;	/* 80 col x 25 line color text */
         extmode[2] = 4;	/* 320 x 200 4-color graphics */
         extmode[3] = 6;	/* 640 x 200 2-color graphics */
         extmode[4] = 13;	/* 320 x 200 16-color graphics */
         extmode[5] = 14;	/* 640 x 200 16-color graphics */
         extmode[6] = 16;	/* 640 x 350 16-color graphics */
         extmode[7] = 17;	/* 640 x 480 2-color graphics */
         extmode[8] = 18;	/* 640 x 480 16-color graphics */
         argc = 10;
         if (color256()) {
            extmode[9] = 19;	/* 320 x 200 256-color graphics */
            argc = 11;
            }
         if (chiptype == TSENG || ET4000) {
            extmode[2] = 34;	/* 132 col x 44 line color text */
            extmode[3] = 35;	/* 132 col x 25 line color text */
            extmode[4] = 4;	/* 320 x 200 4-color graphics */
            extmode[5] = 6;	/* 640 x 200 2-color graphics */
            extmode[6] = 13;	/* 320 x 200 16-color graphics */
            extmode[7] = 14;	/* 640 x 200 16-color graphics */
            extmode[8] = 16;	/* 640 x 350 16-color graphics */
            extmode[9] = 17;	/* 640 x 480 2-color graphics */
            extmode[10] = 18;	/* 640 x 480 16-color graphics */
            argc = 12;
            if ((peek(0x489,0) & 0x20) == 0) {
               if (color256()) {
                  extmode[11] = 19;/* 320 x 200 256-color graphics */
                  argc = 13;
                  if (is512k) {
                     extmode[12] = 46;/* 640 x 480 256-color graphics */
                     argc = 14;
                     if (inp(0x3C2) & 0x60 == 0) {
                        extmode[13] = 55;
                        argc = 15;
                        } 
                     }
                  }
               }
            else {
               extmode[11] = 106;	/* 800 x 600 16-color graphics */
               argc = 13;
               }
            }
         }
      else if (strcmp(argv[1], "1024") == 0) {
         if (chiptype < CHIPS451) {
            printf("Cannot display 1024x768 resolution on this video adapter\n");
            goto usage;
            }
         if (is512k == 0) {
            printf("Cannot display 1024x768 resolution without 512K bytes of video memory\n");
            goto usage;
            }
         extmode[0] = 0;	/* 40 col x 25 line color text */
         extmode[1] = 3;	/* 80 col x 25 line color text */
         extmode[2] = 34;	/* 132 col x 44 line color text */
         extmode[3] = 35;	/* 132 col x 25 line color text */
         extmode[4] = 4;	/* 320 x 200 4-color graphics */
         extmode[5] = 6;	/* 640 x 200 2-color graphics */
         extmode[6] = 13;	/* 320 x 200 16-color graphics */
         extmode[7] = 14;	/* 640 x 200 16-color graphics */
         extmode[8] = 16;	/* 640 x 350 16-color graphics */
         extmode[9] = 17;	/* 640 x 480 2-color graphics */
         extmode[10] = 18;	/* 640 x 480 16-color graphics */
         extmode[11] = 55;	/* 1024 x 768 16-color graphics */
         argc = 13;
         if (color256()) {
            extmode[12] = 19;	/* 320 x 200 256-color graphics */
            extmode[13] = 46;	/* 640 x 480 256-color graphics */
            argc = 15;
            }
         }
      else if (strcmp(argv[1], "msync") == 0) {
         extmode[0] = 0;	/* 40 col x 25 line color text */
         extmode[1] = 3;	/* 80 col x 25 line color text */
         extmode[2] = 34;	/* 132 col x 44 line color text */
         extmode[3] = 35;	/* 132 col x 25 line color text */
         extmode[4] = 4;	/* 320 x 200 4-color graphics */
         extmode[5] = 6;	/* 640 x 200 2-color graphics */
         extmode[6] = 13;	/* 320 x 200 16-color graphics */
         extmode[7] = 14;	/* 640 x 200 16-color graphics */
         extmode[8] = 16;	/* 640 x 350 16-color graphics */
         extmode[9] = 17;	/* 640 x 480 2-color graphics */
         extmode[10] = 18;	/* 640 x 480 16-color graphics */
         extmode[11] = 41;	/* 800 x 600 16-color graphics */
         argc = 13;
         if (color256()) {
            extmode[12] = 19;	/* 320 x 200 256-color graphics */
            argc = 14;
            }
         }
#endif
      else if (strcmp(argv[1], "ega") == 0 || strcmp(argv[1], "EGA") == 0) {
default_ega:
         extmode[0] = 0;	/* 40 col x 25 line color text */
         extmode[1] = 3;	/* 80 col x 25 line color text */
         extmode[2] = 4;	/* 320 x 200 4-color graphics */
         extmode[3] = 6;	/* 640 x 200 2-color graphics */
         extmode[4] = 13;	/* 320 x 200 16-color graphics */
         extmode[5] = 14;	/* 640 x 200 16-color graphics */
         extmode[6] = 16;	/* 640 x 350 16-color graphics */
         argc = 8;
         if (mode_tbl[12].pixels >= 720) {
            extmode[7] = 12;	/* 832 x 350 16-color graphics */
            if (color256() && chiptype == TSENG)
               extmode[7] = 45;	/* 640 x 350 64-color graphics */
            argc = 9;
            }
         }
      else
         goto single;
      }
   else if(argc < 18) {		/* a list of pre-programmed modes */
single:
      for(i=1; i<argc; i++) { /* record parmeters in decimal */
	 extmode[i-1] = atoi(argv[i]);
         if (extmode[i-1] >= 128)
            goto usage;
         }
      }

   else {			/* Cannot specify parameters! */
      printf ("Invalid demo request\n");
      exit(1);
      }
do_demo:
   for(i=1; i<argc; i++) {
      vmode = extmode[i-1];
      setup = &mode_tbl[vmode];
      if (setup->frames == 0)
         continue;

      init();
      if(setup->text == 0)
         graph_test();
      else
         text_test();

      pwait();
      if (lflag) {
         if (i == argc-1)
            i = 0;
         }
      }
   vidmode(monostate? 7: 3);
   /*
   vidmode(vmode+0x80);
   pokeb(0x487, 0, peek(0x487 ,0) & 0x7F);
   */
   }

pwait()
   {
   if (lflag) {
      delay(2000);
      }
   else
      while(waitkey() == 0)
         ;
   }

atoh(s)
   char s[];
   {
   unsigned val;

   val = 0;
   sscanf(s,"%x", &val);
   return val;
   }

delay(time)    /* wait 0-10000 milliseconds */
   int time;
   {
   int i, t1, t2;
   int millis, secs;

   if((time <= 0) || (time > 10000))
      time = 1000;

   getdt();
   millis = msecs;
   secs = seconds; /* get entry time */

   while(1) {
      getdt();

      if(waitkey())
         return (-1);

      if(seconds == 0)
         seconds = 60;
      t2 = seconds - secs;
      t2 *= 1000;

      t1 = t2 + msecs - millis;
      if(t1 < 0)
         t1 += 1000;
      if(t1 > time)
         return 0;
      }
   }

text_test()
   {
   char *t_type;
   int i;
   char s[256];

#ifdef CITIZEN
   t_type = "Colour";
#else
   t_type = "Color";
#endif
   if(regen == 0xB000) t_type = "Monochrome";
   sprintf(s, "%dx%d %s Text", textcols, textrows, t_type);
   color_s(0, (textcols - 12)/2, 0x07, s);

   text_line(23);

   if((crtc == 0x3B4) || (textcols < 80) || (textcols > 128))
      char_set(1);
   else {
      color_bar(1);
      if (textrows > 42)
         color_bar(26);
      }

   attributes(17);

   color_s(0,0, 0x07, "CURSOR< >");
   outp(crtc, 14);
   outp(crtc+1, 0);
   outp(crtc, 15);
   outp(crtc+1, 7);
   }

graph_test()
   {
   unsigned width, height;
   unsigned x1, x2, xinc, y1, y2, yinc;
   int palette, i, j;
   int offset;
   char s[256];
   char *cptr;

#ifdef CITIZEN
   sprintf(s, "%d Colour %dx%d Graphics", zmax, xmax, ymax == 592? 600: ymax);
#else
   sprintf(s, "%d Color %dx%d Graphics", zmax, xmax, ymax);
#endif
   color_s (1, (xmax/8 - 26)/2, 0x07, s);
/*
 *	Border box (square corners)
 */
   set_color(1);
   box (0, 0, xmax-1, ymax-1);
/*
 *	Video mode information in lower left corner
 */
   if (zmax <= 16) {
      if (textcols > 40)
         sprintf(s, "Video mode %d: 8x%d char. cell", vmode, text_height);
      else
         sprintf(s, "Mode %d", vmode);
      color_s(textrows-1, i = textcols-strlen(s), 0x1A, s);
      box (i = 8*i-1, j = text_height*(textrows-1)-1, xmax-1, ymax-1);
      --i;
      line (i, j, i, ymax-1);
      }
/*
 *	Filled rectangle
 */
   if (bits_per_pixel < 8) {
      set_color(6);
      custom_box();
      }
   else {
      if (digital_flag == 0) {
         pal_tab[0] = pal_tab[1] = pal_tab[2] = 0;
         for (i=0, cptr = &pal_tab[3]; i<178*64; i += 64) {
            *cptr++ = i/178;
            *cptr++ = 0;
            *cptr++ = 0;
            }
         set_palette(78, 256-78, &pal_tab[3]);
         }
      custom_256();
      }
/*
 *	Christmas tree figure
 */
   xinc = 5 * xmax;
   xinc /= 3 * ymax;
   if (ymax <= 350)
      xinc /= 2;
   if (xinc <= 0)
      xinc = 1;
   palette = 0;
   for(i=0, j=0; i<ymax/4; i++, j += xinc) { /* h_line test */
      set_color(palette = (palette+1) & 0x3F);
      if (ymax > 350) {
         if(i&2) {
            line(xmax/2, ymax/8+i, xmax/2+j, ymax/8+i);
            i++;
            line(xmax/2, ymax/8+i, xmax/2+j, ymax/8+i);
            }
         else {
            line(xmax/2, ymax/8+i, xmax/2-j, ymax/8+i);
            i++;
            line(xmax/2, ymax/8+i, xmax/2-j, ymax/8+i);
            }
         }
      else {
         if(i&1)
            line(xmax/2, ymax/8+i, xmax/2+j, ymax/8+i);
         else
            line(xmax/2, ymax/8+i, xmax/2-j, ymax/8+i);
         }
      }
/*
 *	Five pointed star figure
 */
   set_color(9);
   x1 = xmax/4;
   y1=ymax/2;
   j = ymax/64;
   i = (j*xmax)/ymax;
   line(x1-63*i/6, y1-14*j/4, x1+63*i/6, y1-14*j/4);
   line(x1+63*i/6, y1-14*j/4, x1-39*i/6, y1+60*j/4);
   line(x1-39*i/6, y1+60*j/4, x1, y1-60*j/4);
   line(x1, y1-60*j/4, x1+39*i/6, y1+60*j/4);
   line(x1+39*i/6, y1+60*j/4, x1-63*i/6, y1-14*j/4);
   set_color(2);
   paint(x1, y1, 1);
/*
 *	Concentric ellipses
 */
#ifdef PAINTED
   j = palette = 1;
#endif
   x1 = (35*xmax) >> 6;
   y1 = (23*ymax) >> 5;
   i = (6 * xmax) >> 5;
   x1 -= 4;
   aspect = ((long)ymax)*384/xmax;
   while (aspect > 8) {
      set_color(palette);
      ellipse(x1, y1, i, aspect);
#ifdef PAINTED
      if (j == 3)
         paint(x1, y1, palette);
      else if (j == 4) {
         set_color(0);
         paint(x1, y1, palette);
         j = 1;
         }
      else if (j == zmax)
         j = 1;
      ++j;
#endif
      if (++palette >= zmax)
         palette = 1;
      aspect -= (aspect >> 2) + 8;
      }
/*
 *	Fishnet on the lower left corner
 */
   aspect = 15 * xmax / ymax;
   aspect /= 2;
   y1 = ymax/3;
   set_color(3);
   line(0, ymax-1, j = xmax/2, ymax-1);
   for (i=ymax-1; j > 0; i -= 10, j -= aspect)
      line(0, i, j, ymax-1);
   line(0, y1, 0, ymax-1);
   custom_msg();
/*
 *  Cycle through palettes for 256 color mode
 */
   if (bits_per_pixel >= 8 && digital_flag == 0) {
      pwait();
      set_palette(78, 256-78, &pal_tab[2]);
      pwait();
      set_palette(78, 256-78, &pal_tab[1]);
      pwait();
      for (cptr = &pal_tab[0]; cptr < pal_tab + sizeof(pal_tab) - 2; cptr += 3)
         cptr[2] = cptr[1] = cptr[0];
      set_palette(78, 256-78, &pal_tab[0]);
      pwait();
      offset = 180/3;
      offset = offset * 3 + 1;
      for (j = 0, i = 0; i <= 180; j += 3, i++) {
         pal_tab[j] = rainbow[i];
         pal_tab[(j+offset)%(180*3)] = rainbow[i];
         pal_tab[(j+offset*2)%(180*3)] = rainbow[i];
         }
      set_palette(78, 256-78, &pal_tab[0]);
      }
   }

color_bar(line)
int line;
{
unsigned color;
unsigned int row, col;
unsigned int bar_height;
unsigned int text_attr;

   bar_height = 16;
   for(color = 0; color < 16; color++) {
      col = color*strlen(bar_pattern);
      text_attr = (color << 4) & 0x70;
      if (color & 0x08)
         text_attr |= 0x0F;
      if (text_attr == 0)
         text_attr = 0x07;
      for(row=line; row<line+bar_height; row++) {
	 color_s(row, col, color, bar_pattern);
         }
      color_s(line+bar_height/2, col+1, text_attr, color_abbrev[color]);
   }
}


attributes(line)
int line;
{
unsigned col;

   col = (textcols - 35) / 2;
   if(setup->blink) color_s(line+1, col, 0x87,"    This line should be BLINKING    ");
   color_s(line+2, col, 0x70, "  This line should be INVERSE VIDEO ");
   color_s(line+3, col, 0x0F, "   This line should be HI-LIGHTED   ");
   if(regen == 0xB000) color_s(line+4, col, 0x01, "   This line should be UNDERLINED   ");
   color_s(line+5, col, 0x00, "   This line should NOT be VISIBLE  ");

}

text_line(line)
   int line;
   {
   unsigned c;

   for(c=0; c<textcols; c++)
      (*color_c)(line, c, 0x07, c+'A');
   }

char_set(line)
   int line;
   {
   unsigned int c, row, col, addr, val;
   unsigned int column;
   unsigned int attr;

   for(row=0; row<16; row++) {	/* print color names */
      attr = (row % 15) + 1;
      color_s(row + line, 0, attr, colors[attr]);
      if(textcols > 120)
	 color_s(row + line, textcols-strlen(colors[attr]), attr, colors[attr]);
      if(textrows>39) {
	 color_s(row + line + 24, 0, attr, colors[attr]);
	 if(textcols > 120)
	    color_s(row + line + 24, textcols-strlen(colors[attr]),
                  attr, colors[attr]);
	 }
      }

   for(column=0; column < (textcols-20); column+=40) {
      for(c=0; c<256; c++) {	   /* draw the char set */
         val = c;
         if (vmode == 0) {	   /* special case, get rid of "funnies" */
            val++;		   /* i.e., "blank" characters and graphics */
            if (val > 31)
               val++;
            if (val > 175)
               val -= 175;
            }
	 row = c / 16;
	 attr = (row % 15) + 1;
	 col = (c % 16) * 2 + 8 + ((row^(column/40))&1);
	 (*color_c)(row + line, col+column, attr, val);
	 if (textrows>42)
	    (*color_c)(row + line + 24, (col+column) ^ 1, attr, val);
	 }
      }
   }

text_char(row, col, attr, c)
   unsigned row, col;
   unsigned char attr;
   unsigned char c;
   {
   pokew(row_start[row]+col+col, regen, (attr << 8) | c);
   }

one_char(row, col, attr, c)
   unsigned int row, col;
   unsigned int attr;
   unsigned int c;
   {
   unsigned i, seg, ofs, bkcolor;
   volatile unsigned char far *ptr;

   ptr = fontptr + c * text_height;
   row *= text_height;
   if (zmax < 4) {
      if (attr & 0x0100) {
         if (attr & 0x00FF)
            for(i=0; i<text_height; i++) {
               bkcolor = peek(row_start[row]+col, regen);
               pokeb(row_start[row++]+col, regen, *ptr++ | bkcolor);
               }
         else
            for(i=0; i<text_height; i++) {
               bkcolor = peek(row_start[row]+col, regen);
               pokeb(row_start[row++]+col, regen, (~*ptr++) & bkcolor);
               }
         }
      else if (attr & 0x0200)
         for(i=0; i<text_height; i++) {
            bkcolor = peek(row_start[row]+col, regen);
            pokeb(row_start[row++]+col, regen, *ptr++ ^ bkcolor);
            }
      else
         for(i=0; i<text_height; i++) {
            pokeb(row_start[row++]+col, regen, *ptr++);
            }
      }
   else {
      (*get_dot)(0, row);		/* Initialize "regen" */
      if ((attr & 0x0200) != 0) {
         wrt_ega(((attr<<8) & 0x0F00) | 0x02, 0x3C4);
         wrt_ega(0x1803, 0x3CE);
         for(i=0; i<text_height; i++)
            pokeb(row_start[row++]+col, regen, *ptr++);
         wrt_ega(0x0003, 0x3CE);
         wrt_ega(0x0F02, 0x3C4);
         return;
         }
      wrt_ega(0x0205, 0x3CE);
      if ((attr & 0x0100) == 0) {
         wrt_ega(0xFF08, 0x3CE);
         bkcolor = (attr >> 4) & 0x0F;
         for (i=0; i<text_height; i++)
            pokeb(row_start[row++]+col, regen, bkcolor);
         row -= text_height;
         }
      for(i=0; i<text_height; i++) {
         wrt_ega((*ptr++ << 8) + 8, 0x3CE);
         pokeb(row_start[row++]+col, regen, attr);
         }
      wrt_ega(0xFF08, 0x3CE);
      wrt_ega(0x0005, 0x3CE);
      }
   }

two_char(row, col, attr, c)
   unsigned int row, col;
   unsigned int attr;
   unsigned int c;
   {
   unsigned i, bkcolor;
   unsigned char far *ptr;

   ptr = fontptr + c * text_height;
   col <<= 1;
   row *= text_height;
   if (attr & 0x0100)
      for(i=0; i<text_height; i++) {
         bkcolor = peek(row_start[row]+col, regen)
               & (~expand_left[*ptr & 0xFF]);
         pokew(row_start[row++]+col, regen, (expand_left[*ptr++]
               & color_table[attr & 0x0F]) | bkcolor);
         }
   else if (attr & 0x0200)
      for(i=0; i<text_height; i++) {
         bkcolor = peek(row_start[row]+col, regen)
               & (~expand_left[*ptr & 0xFF]);
         pokew(row_start[row++]+col, regen, (expand_left[*ptr++]
               & color_table[attr & 0x0F]) ^ bkcolor);
         }
   else
      for(i=0; i<text_height; i++) {
         pokew(row_start[row++]+col, regen, expand_left[*ptr++]
               & color_table[attr & 0x0F]);
         }
   }

four_char(row, col, attr, c)
   unsigned int row, col;
   unsigned int attr;
   unsigned int c;
   {
   unsigned int i, bkcolor;
   unsigned char far *ptr;

   ptr = fontptr + c * text_height;
   col <<= 2;
   row *= text_height;
   if (attr & 0x0100)
      for(i=0; i<text_height; i++) {
         bkcolor = peek(row_start[row]+col, regen) &
               (~expand_left[c = *ptr++]
               & color_table[attr & 0x0F]);
         pokew(row_start[row]+col, regen, (expand_left[c]
               & color_table[attr & 0x0F]) | bkcolor);
         bkcolor = peek(row_start[row]+col+2, regen) & (~expand_right[c]);
         pokew(row_start[row++]+col+2, regen, (expand_right[c]
               & color_table[attr & 0x0F]) | bkcolor);
         }
   else if (attr & 0x0200)
      for(i=0; i<text_height; i++) {
         bkcolor = peek(row_start[row]+col, regen) &
               (~expand_left[c = *ptr++]
               & color_table[attr & 0x0F]);
         pokew(row_start[row]+col, regen, (expand_left[c]
               & color_table[attr & 0x0F]) | bkcolor);
         bkcolor = peek(row_start[row]+col+2, regen) & (~expand_right[c]);
         pokew(row_start[row++]+col+2, regen, (expand_right[c]
               & color_table[attr & 0x0F]) ^ bkcolor);
         }
   else
      for(i=0; i<text_height; i++) {
         pokew(row_start[row]+col, regen, expand_left[c = *ptr++]
               & color_table[attr & 0x0F]);
         pokew(row_start[row++]+col+2, regen, expand_right[c]
               & color_table[attr & 0x0F]);
         }
   }

eight_char(row, col, attr, c)
   unsigned int row, col, attr, c;
   {
   unsigned int i, j, k;
   int bkcolor;
   volatile unsigned char far *ptr;
   unsigned int test;

   ptr = fontptr + c * text_height;
   col <<= 3;
   row *= text_height;
   if (chiptype == TSENG || chiptype == ET4000) {
      outp(0x3CD, bank_sel[row]);
      test = row_start[row];
      }
   if (attr & 0x0100)
      for(i=0; i<text_height; i++, ptr++, row++) {
         for (j=0x80, k=0; j; k++, j >>= 1)
            if (*ptr & j) {
               if (test > row_start[row]+col+k) {
                  test = row_start[row]+col+k;
                  if (chiptype == TSENG)
                     outp(0x3CD, inp(0x3CD) + 9);
                  else if (chiptype == ET4000)
                     outp(0x3CD, inp(0x3CD) + 0x11);
                  }
               pokeb(row_start[row]+col+k, regen, attr);
               }
         }
   else if (attr & 0x0200)
      for(i=0; i<text_height; i++, ptr++, row++) {
         for (j=0x80, k=0; j; k++, j >>= 1)
            if (*ptr & j) {
               if ((test ^ row_start[row]+col+k) & 0x8000) {
                  test = row_start[row]+col+k;
                  if (chiptype == TSENG)
                     outp(0x3CD, inp(0x3CD) + 9);
                  else if (chiptype == ET4000)
                     outp(0x3CD, inp(0x3CD) + 0x11);
                  }
               bkcolor = peek(row_start[row]+col+k, regen);
               pokeb(row_start[row]+col+k, regen, attr ^ bkcolor);
               }
         }
   else
      for(i=0; i<text_height; i++, ptr++, row++) {
         for (j=0x80, k=0; j; k++, j >>= 1) {
               if ((test ^ row_start[row]+col+k) & 0x8000) {
                  test = row_start[row]+col+k;
                  if (chiptype == TSENG)
                     outp(0x3CD, inp(0x3CD) + 9);
                  else if (chiptype == ET4000)
                     outp(0x3CD, inp(0x3CD) + 0x11);
                  }
            pokeb(row_start[row]+col+k, regen, (*ptr & j)?attr: 0);
            }
         }
   }

eight_get(x, y)
   unsigned int x, y;
   {
   if (chiptype == TSENG)
      outp(0x3CD, bank_sel[y] + (row_start[y]+(long)x > 0x10000L? 9: 0));
   else if (chiptype == ET4000)
      outp(0x3CD, bank_sel[y] + (row_start[y]+(long)x > 0x10000L? 0x11: 0));
   return (peek(row_start[y]+x, regen) & 0xFF);
   }

eightcolor(x, y, testcolor)
   unsigned int x, y;
   unsigned int testcolor;
   {
   if (chiptype == TSENG)
      outp(0x3CD, bank_sel[y] + (row_start[y]+(long)x > 0x10000L? 9: 0));
   else if (chiptype == ET4000)
      outp(0x3CD, bank_sel[y] + (row_start[y]+(long)x > 0x10000L? 0x11: 0));
   return ((peek(row_start[y] + x, regen) & 0xFF) - testcolor);
   }

eight_bit(col, row)
   unsigned int col, row;
   {
   if (chiptype == TSENG)
      outp(0x3CD, bank_sel[row] + (row_start[row]+(long)col > 0x10000L? 9: 0));
   else if (chiptype == ET4000)
      outp(0x3CD, bank_sel[row] + (row_start[row]+(long)col > 0x10000L? 0x11: 0));
   pokeb(row_start[row] + col, regen, color);
   }

ibm_char(row, col, attr, c)
   unsigned int row, col, attr, c;
   {
   unsigned int i, j, k;
   int bkcolor;
   volatile unsigned char far *ptr;
   unsigned int test;

   ptr = fontptr + c * text_height;
   row *= text_height;
   col <<= 1;
   if (attr & 0x0100)
      for(i=0; i<text_height; i++, ptr++, row++) {
         test = row_start[row] + col;
         for (j=0x0888; j & 0xFF00; j >>= 1) {
            wrt_ega(ibm_map[j & 0x0F], 0x3C4);
            if (*ptr & j & 0x00F0) {
               pokeb(test, regen, attr);
               }
            if (*ptr & j & 0x000F) {
               pokeb(test+1, regen, attr);
               }
            }
         }
   else if (attr & 0x0200)
      for(i=0; i<text_height; i++, ptr++, row++) {
         test = row_start[row] + col;
         for (j=0x0888; j & 0xFF00; j >>= 1) {
            wrt_ega(ibm_map[j & 0x0F], 0x3C4);
            bkcolor = peek(test, regen);
            if (*ptr & j & 0x00F0) {
               pokeb(test, regen, attr ^ bkcolor);
               }
            if (*ptr & j & 0x000F) {
               pokeb(test+1, regen, attr ^ (bkcolor >> 8));
               }
            }
         }
   else
      for(i=0; i<text_height; i++, ptr++, row++) {
         test = row_start[row] + col;
         for (j=0x0888; j & 0xFF00; j >>= 1) {
            wrt_ega(ibm_map[j & 0x0F], 0x3C4);
            pokeb(test, regen, (*ptr & j & 0x00F0)? attr: 0);
            pokeb(test+1, regen, (*ptr & j & 0x000F)? attr: 0);
            }
         }
   }

color_s(row, col, attr, s)
   unsigned row, col, attr;
   unsigned char s[];
   {
   int i;
   int c;

   for(i=0; (c = s[i]) != 0; i++) {
      (*color_c)(row, col+i, attr, c);
      }
   }

set_color(c)
   int c;
   {
   if (zmax <= 16)
      color = color_table[c & 0x0F];
   else
      color = c % zmax;
   return color;
   }

init()
   {
   unsigned int i, j;
   unsigned int row;
   unsigned int c, mask1, mask2;

   xmax      = setup->pixels;
   ymax      = setup->lines;
   max_scan  = setup->frames;
   textrows  = setup->rows;
   textcols  = setup->chars;

   regen     = setup->segmt;
   memend    = setup->size;
   crtc      = setup->addr;

   zmax = setup->colors;

   row_start[0] = 0x55;
   if(setup->text == 0) {
      text_height = setup->charsize;
      bits_per_pixel = setup->bits;
      fontptr = getfont(text_height < 13? 3: text_height < 15? 2: 6);

      if(bits_per_pixel == 1) {
         if (zmax <= 2) {
            dot = one_bit; 	       /* draw_dot function */
            get_dot = one_get;	       /* read_dot function */
            find_color = onecolor;
            color_table = &one_color[0];  /* color lookup/expand table */
            }
         else {
	    if (chiptype == ET4000) {
               dot = et4000_bit; 	       /* draw_dot function */
               get_dot = et4000_get;	       /* read_dot function */
               find_color = et4000color;
 	       limit1 = limit2 = 0xFFFF;
	       j = xmax/8;
	       for(row = i = 0; i < 1024; ++i) {	/* for each scanline */
		  if (row+j < 0 && limit1 == 0xFFFF) {
		     limit1 = i+1;
		     }
		  row_start[i] = row;	/* calc the row start address */
		  row += j;		/* base addr of group of lines */
		  }
	       }
	    else {
               dot = ega_bit; 	       /* draw_dot function */
               get_dot = ega_get;	       /* read_dot function */
               find_color = egacolor;
 	       limit1 = limit2 = 0xFFFF;
	       j = xmax/8;
	       for(row = i = 0; i < 1024; ++i) {	/* for each scanline */
		  if (row >= 0xE000 && limit1 == 0xFFFF) {
		     limit1 = i;
		     row -= 0xE000;
		     }
		  row_start[i] = row;	/* calc the row start address */
		  row += j;		/* base addr of group of lines */
		  }
               }
	    color_table = &ega_color[0];
            }
         color_c = one_char;	       /* text output routine */
         get_shift = &one_shift[0];    /* read_dot alignment shift count table */
         clear_word = 0;	       /* buffer clearing value */
         pixels_per_byte = 8;
         }
      else if(bits_per_pixel == 2) {
	 pixels_per_byte = 4;
	 dot = two_bit;
	 color_table = &two_color[0];
	 get_dot = two_get;
	 find_color = twocolor;
	 get_shift = &two_shift[0];
	 color_c = two_char;
	 clear_word = 0;
	 for(c=0; c<256; c++) { /* build the text bit pattern expand table */
	    expand_left[c] = 0;
	    mask1 = 0x08;
	    for(i=0xC000; i; i>>=2) { /* color bit mask */
	       if(c & mask1)
		  expand_left[c] |= i;
	       mask1 >>= 1;
	       if(mask1 == 0)
		  mask1 = 0x80;
	       }
	    }
         }
      else if (bits_per_pixel == 4) {  /* 4 bits per pixel */
	 pixels_per_byte = 2;
	 dot = four_bit;
	 color_table = &four_color[0];
	 get_dot = four_get;
	 find_color = fourcolor;
	 get_shift = &four_shift[0];
	 color_c = four_char;
	 clear_word = 0;
	 for(c=0; c<256; c++) { /* build the text bit pattern expand table */
	    expand_left[c] = expand_right[c] = 0;
	    mask1 = 0x20;
	    mask2 = 0x02;
	    for(i=0xF000; i; i>>=4) { /* color bit mask */
	       if(c & mask1) expand_left[c] |= i;
	       mask1 >>= 1;
	       if(mask1 == 0x08) mask1 = 0x80;

	       if(c & mask2) expand_right[c] |= i;
	       mask2 >>= 1;
	       if(mask2 == 0) mask2 = 0x08;
	       }
            }
         }
      else if (bits_per_pixel == 8) {
	 pixels_per_byte = 1;
         if (chiptype == ET4000) {
            for(row = i = 0, j=0x00; i < 1024; i++) {	/* for each VTOTAL */
               bank_sel[i] = j;		/* calc the row start address */
               if (row > 0xF800 && (row + xmax) < 0x0800)
                  j += 0x11;
               row += xmax;		/* base addr of group of lines */
               }
            }
         else {
            for(row = i = 0, j=0x40; i < 1024; i++) {	/* for each VTOTAL */
               bank_sel[i] = j;		/* calc the row start address */
               if (row > 0xF800 && (row + xmax) < 0x0800)
                  j += 0x09;
               row += xmax;		/* base addr of group of lines */
               }
            }
         dot = eight_bit;
         get_dot = eight_get;
         find_color = eightcolor;
	 color_c = eight_char;
	 clear_word = 0;
         }
      else {				/* Cirrus/IBM 256 color modes */
	 pixels_per_byte = 4;
	 dot = ibm_bit;
	 get_dot = ibm_get;
	 find_color = ibmcolor;
	 color_c = ibm_char;
	 clear_word = 0;
         }

      row = 0;
      if (row_start[0] == 0x55)
         for(i = 0; i < 1024; i += max_scan) { /* for each VTOTAL */
            for(j = 0; j < max_scan; j++)      /* for each scan row */
               row_start[i+j] = row + (j<<13); /* calc the row start address */
            row += xmax/pixels_per_byte;       /* base addr of group of lines */
            }

      aspect = ((unsigned)ymax * 80) / xmax;
           /*  ((ymax * 5) * 64) / (xmax * 4)  */
      if(aspect > 63)
         aspect = 63;
      }
   else { /* TEXT MODE SETUPS */
      for(i=0; i<1024; i++) { /* for each VTOTAL */
	 row_start[i] = i * textcols * 2; /* calc the row start address */
         }
      color_c = text_char;
      clear_word = 0x0720;
      }
   vidmode(vmode);
   }

/*
 *
 * BOX: Draw a (color) rectangle from upper left = (x1,y1)
 *		to lower right = (x2,y2)
 *
 */
box( x1, y1, x2, y2 )
   int x1, y1, x2, y2;
   {
   line(x1,y1,x2,y1);
   line(x2,y1,x2,y2);
   line(x2,y2,x1,y2);
   line(x1,y2,x1,y1);
   }

filled_box(x1, y1, x2, y2)
   unsigned int x1,y1, x2,y2;
   {
   unsigned int t;

   if(y2<y1) {	/* draw box top to bottom */
      t=y2; y2=y1;  y1=t;
      }
   while (y1<=y2) {     /* using horizontal lines	*/
      line(x1, y1, x2, y1);
      y1++;
      }
   }

shaded_box(x1, y1, x2, y2, color1)
   unsigned int x1,y1, x2,y2, color1;
   {
   unsigned int t, colorused;
   int change;

   colorused = color1 - 1;
   change = 1;
   if(y2<y1) {	/* draw box top to bottom */
      t=y2; y2=y1; y1=t;
      }
   while (y1<=y2) {     /* using horizontal lines	*/
      if ((colorused = (colorused + change) & 0xFF) < color1) {
         change = - change;
         continue;
         }
      set_color(colorused);
      line(x1, y1, x2, y1);
      y1++;
      if (digital_flag) {
         if (y1 <= y2) {
            line(x1, y1, x2, y1);
            y1++;
            }
         if (y1 <= y2) {
            line(x1, y1, x2, y1);
            y1++;
            }
         }
      }
   }

/********************************************************************
 *
 * PAINT: Fill in the convex area containing (x,y) bounded by color (bound)
 *		with color set by last call to set_color().
 *
 *******************************************************************/
#ifndef FILL
paint (xi, yi, bound)
   int xi, yi, bound;
   {
#ifdef NOFILL
   int yinc;

   if ((*get_dot)(xi, yi) == bound)
      return;
   yinc = -1;
   pass_paint(xi, yi, bound, yinc);
   yinc = 1;
   pass_paint(xi, yi+1, bound, yinc);
   }

pass_paint (xi, yi, bound, yinc)
   int xi, yi, bound, yinc;
   {
   int xl, xk, xn, xp, yp;

   yp = yi;
   xp = xi + 1;
   xl = xi - 1;
   if ((*find_color)(xi, yp, bound) == 0) {
      do {
         xl -= pixels_per_byte;
         if (xl < 0) {
            xl = 0;
            break;
            }
         } while ((*find_color)(xl, yp, bound) == 0);
      xl += pixels_per_byte;
      do {
         xp += pixels_per_byte;
         if (xp > xmax) {
            xp = xmax;
            break;
            }
         } while ((*find_color)(xp, yp, bound) == 0);
      xp -= pixels_per_byte;
      }
   while ((*get_dot)(xl, yp) != bound)
      if (--xl < 0)
         break;
   xl++;
   while ((*get_dot)(xp, yp) != bound)
      if (++xp > xmax)
         break;
   xp--;

   while (1) {
      line(xl, yp, xp, yp);
      yp += yinc;
      if (yp >= ymax || yp < 0)
         return;
#ifdef NONCONVEX
      while ((*get_dot)(xl, yp) == bound) {
         /* loss on the left */
         if (++xl > xp)
            return;
         }
      while ((*get_dot)(xl - 1, yp) != bound) {
         /* growth to left */
         if (--xl == 0)
            break;
         }
      while ((*get_dot)(xp, yp) == bound) {
         /* loss on the right */
         --xp;
         }
      while ((*get_dot)(xp + 1, yp) != bound) {
         /* growth to right */
         if (++xp == xmax)
            break;
         }
      pass_loop:
      xk = (xl + pixels_per_byte - 1)&(- pixels_per_byte);
      if (xk >= xp)
         xk = xp;
      for (xn = xl; xn < xk; xn++) {
         if ((*get_dot)(xn, yp) == bound) {
            printf("Recursive Paint Call (%d-%d, %d)\n", xl, xn-1, yp);
            pass_paint(xl, yp, bound, yinc);
            while ((*get_dot)(xn, yp) == bound)
               xn++;
            xl = xn;
            }
         }
      for (xn = xk; xn < xp; xn += pixels_per_byte) {
         if ((*find_color)(xn, yp, bound)) {
            while ((*get_dot)(xn, yp) != bound)
               xn++;
            xn--;
            if (xn < xk) {
               printf("Recursive Paint Call (%d-%d, %d)\n", xl, xn-1, yp);
               pass_paint(xl, yp, bound, yinc);
               while ((*get_dot)(xn, yp) == bound)
                  xn++;
               xl = xn;
               goto pass_loop;
               }
            break;
            }
         }
#else
      if ((*get_dot)(xl, yp) != bound) {
         do {
            xl--;
            } while(xl >= 0 && (*get_dot)(xl, yp) != bound);
         xl++;
         }
      else
         do {
            xl++;
            } while(xl < xp && (*get_dot)(xl, yp) == bound);
      if ((*get_dot)(xp, yp) == bound)
         do {
            xp--;
            } while(xp >= xl && (*get_dot)(xp, yp) == bound);
      else {
         do {
            xp++;
            } while(xp < xmax && (*get_dot)(xp, yp) != bound);
         xp--;
         }
      if (xl > xp)
         break;
#endif
      }
#endif
   }
#else
paint (xi, yi, bound)
   int xi, yi, bound;
   {
   fill(xi, yi, color);
   }
#endif
ellipse(x, y, major, aspect)
    int x, y, major;
    unsigned int aspect;          /* aspect ratio * 256 */
    {
    int row, col, two_x, two_y;
    long d, cincr, rincr, two_a, two_b, alpha, beta;

    d = ((long)major)*((long)aspect);
/* the following code applies if "major" is the major axis length */
#ifdef REALMAJOR
    if (aspect < 256) {
        two_a = ((long)major)*((long)major)*2L;
        two_b = ((d >> 3) * (d >> 3)) >> 9;
        d = (d+64) >> 7;
        }
    else {
        two_a = ((d >> 4) * (d >> 4)) >> 7;
        two_b = ((long)major)*((long)major)*2L;
        d = major * 2;
        }
#else
/* the following code applies if "major" is the x-axis length */
    two_a = ((long)major)*((long)major)*2L;
    two_b = ((d >> 4) * (d >> 4)) >> 7;
    d = (d+64) >> 7;
#endif
    row = y + (d >> 1);
    col = x;
    two_x = x << 1;
    two_y = y << 1;
    alpha = two_a >> 1;
    beta = two_b >> 1;
    rincr = - (alpha * d);
    cincr = beta;
    d = ((beta - alpha) << 1) + (alpha >> 1);
/*
 *  For the remaining pare of the routine, only 'col', 'row', 'two_y',
 *  'two_x', 'color', 'two_a', 'two_b', 'cincr', 'rincr' and 'd' are
 *  use.
 */
    while (cincr + rincr < 0) {
        (*dot)(col, row);
        (*dot)(col, two_y - row);
        (*dot)(two_x - col, row);
        (*dot)(two_x - col, two_y - row);
        d += (cincr += two_b);
        col++;
        if (d >= 0) {
            d += (rincr += two_a);
            row--;
            }
        }
    while (row >= y) {
        (*dot)(col, row);
        (*dot)(col, two_y - row);
        (*dot)(two_x - col, row);
        (*dot)(two_x - col, two_y - row);
        if (d <= 0) {
            d += (cincr += two_b);
            col++;
            }
        row--;
	d += (rincr += two_a);
        }
    }

line(x1, y1, x2, y2)
   int x1, y1, x2, y2;
   {
   int i1, i2, d1, d2, d;
   unsigned addr;
   int left, right, x3, x4;

   if (dot == ega_bit || dot == et4000_bit)
      (*dot)(x1, y1);
   d2 = y2 - y1;
   d1 = x2 - x1;
   if(d2 == 0) {    /* horizontal line */
      if(d1 < 0) {  /* set drawing direction */
	 left = x2;
         right = x1;
         }
      else {
	 left = x1;
         right = x2;
         }

      if(setup->bits & 0x20) {   /* quick and dirty lines for now */
         addr = row_start[y1];
         x1 = left;
         x2 = right;
         x3 = addr + (left+3)/4;
         x4 = addr + (right-3)/4;
         left = (left+3) & ~3;
         right = (right-3) | 3;
         color |= color << 8;
         if (left <= right) {
            while (x1 <= left)
               (*dot)(x1++, y1);
            wrt_ega(0xFF08, 0x3CE);
            wrt_ega(0x0F02, 0x3C4);
            filler(x3, regen, x4, color);
            while (right <= x2)
               (*dot)(right++, y1);
            }
         else {
            while(x1 <= x2)
               (*dot)(x1++, y1);
            }
	 return;
         }

      if (bits_per_pixel == 8) {
         addr = row_start[y1];
         left += addr;
         right += addr;
         color |= color << 8;
         if (chiptype == TSENG)
            outp(0x3CD, bank_sel[y1] + (addr > left? 9: 0));
         else if (chiptype == ET4000)
            outp(0x3CD, bank_sel[y1] + (addr > left? 0x11: 0));
         if ((unsigned)left > (unsigned)right) {
            filler(left, regen, 0xFFFF, color);
            if (chiptype == TSENG)
               outp(0x3CD, inp(0x3CD) + 9);
            else if (chiptype == ET4000)
               outp(0x3CD, inp(0x3CD) + 0x11);
            filler(0, regen, right, color);
            }
         else {
            filler(left, regen, right, color);
            }
         return;
         }
      if(left+16 >= right) {   /* quickest way to draw short lines */
	 for(x1 = left; x1<= right; x1++)
            (*dot)(x1, y1);
	 return;
         }

      x1 = left;
      x4 = right;
      if(bits_per_pixel == 1) {
	 x2 = (left  + 7) & 0xFFF8;
	 x3 = (right - 7) & 0xFFF8;
	 left = x2 / 8;
         right = x3 / 8 - 1;
         }
      else if(bits_per_pixel == 2) {
	 x2 = (left  + 3) & 0xFFFC;
	 x3 = (right - 3) & 0xFFFC;
	 left = x2 / 4;
         right = x3 / 4 - 1;
         }
      else if(bits_per_pixel == 4) {
	 x2 = (left  + 1) & 0xFFFE;
	 x3 = (right - 1) & 0xFFFE;
	 left = x2 / 2;
         right = x3 / 2 - 1;
         }

      addr = row_start[y1];
      if(left <= right) {
         if (dot == ega_bit || dot == et4000_bit) {
            wrt_ega(0x0205,0x3CE);
            wrt_ega(0xFF08,0x3CE);
            }
         filler(addr + left, regen, addr + right,
               color | (color << 8)); /* middle of line */
         }
      if(x1 < x2)
         for( ; x1 < x2; x1++)
            (*dot)(x1, y1); /* left edge */
      if(x3 <= x4)
         for( ; x3 <= x4; x3++)
            (*dot)(x3, y1); /* right edge */
      return;
      }

   i1 = 1;
   if (d1 < 0) {
      d1 = -d1;
      i1 = -1;
      }
   i2 = 1;
   if (d2 < 0) {
      d2 = -d2;
      i2 = -1;
      }

   if(d1 > d2) {
      d = d2 + d2 - d1;
      while(1) {
	 (*dot)(x1, y1);
	 if(x1 == x2)
            break;
	 if(d >= 0) {
	    d -= (d1 + d1);
	    y1 += i2;
            }
	 d += d2 + d2;
	 x1 += i1;
         }
      }
   else {
      d = d1 + d1 - d2;
      while (1) {
	 (*dot)(x1, y1);
	 if(y1 == y2)
            break;
	 if(d >= 0) {
	    d -= (d2 + d2);
	    x1 += i1;
	 }
	 d += d1 + d1;
	 y1 += i2;
      }
   }
}

getdt()
   {  /* Read system clock - MS DOS Version */
   union REGS reg;

   reg.h.ah = 0x2A;	      /* Get DATE function */
   intdos(&reg, &reg);
   weekday = reg.h.al;
   year = reg.x.cx;
   month = reg.h.dh;
   day = reg.h.dl;

   reg.h.ah = 0x2C;	      /* Get the TIME function */
   intdos(&reg, &reg);
   hours = reg.h.ch;
   minutes = reg.h.cl;
   seconds = reg.h.dh;
   msecs = reg.h.dl * 10;
   }

waitkey()
   {
   if (kbhit() == 0)
      return 0;
   return (0x100 | getch());
   }

sine_tab(len, max, table)
   int len, max;
   int *table;
   {
   int col;
   long two_a, two_b, d;
   long a, b, aincr, bincr;
   int center, adj;

   col = 0;
   center = len/2;
   a = ((long)len)*((long)len);
   two_a = a+a;
   b = ((long)max)*((long)max);
   two_b = b+b;
   d = 210L * max;
   aincr = -a*d/len;
   bincr = b*d/len;
   d = ((b - a) << 1) + (a >> 1);

   while (bincr + aincr < 0 && col <= center) {
      table[center+col] = max;
      table[center-col] = max;
      d += (bincr += two_b);
      col++;
      if (d >= 0) {
         d += (aincr += two_a);
         max--;
         }
      }
   while (col <= center) {
      table[center+col] = max;
      table[center-col] = max;
      if (d <= 0) {
         d += (bincr += two_b);
         col++;
         }
      max--;
      d += (aincr += two_a);
      }
   }

chipset()
   {
   int temp;
   int save;
   int type;
   int crtcaddr;

   type = EGA;
   crtcaddr = peek(0x463,0);
   temp = rr(crtcaddr, 0xFF);
   outp(crtcaddr+1, 0xC0 ^ temp);
   if (((rr(crtcaddr, 0xFF)^temp) ^ 0xC0) == 0 && rr(crtcaddr, 0xF7) == 0 &&
         rr(crtcaddr, 0xF8) == 0) {
      outp(crtcaddr, 0xFF);
      outp(crtcaddr+1, temp);
      temp = rr(0x3CE, 0xF7);
      outp(0x3CF, 0xFF ^ temp);
      if ((temp ^ rr(0x3CE, 0xF7) ^ 0xFF) == 0) {
         outp(0x3CF, temp);
         type = CHIPS441;
         }
      else {
         outp(0x3CF, temp);
         type = CHIPS435;
         }
      }
   else {
      outp(crtcaddr, 0xFF);
      outp(crtcaddr+1, temp);
      temp = rr(crtcaddr, 0x16);
      outp(crtcaddr+1, 0xFF ^ temp);
      if ((rr(crtcaddr, 0x16) ^ temp ^ 0xFF) == 0) {
         outp(crtcaddr, 0x16);
         outp(crtcaddr+1, temp);
         if (((temp=rr(crtcaddr, 0x23)) & 0x03) == 0) {
            outp(crtcaddr+1, 0x83 | temp);
            if ((rr(crtcaddr, 0x23) & 0x83) == 0x83) {
               outp(crtcaddr, 0x23);
               outp(crtcaddr+1, temp);
               type = TSENG;
               }
            else {
               outp(crtcaddr, 0x23);
               outp(crtcaddr+1, temp);
               goto ibm;
               }
            }
         else {
            ibm:
            type = VGA;
            outp(0x3BF, 0x03);
            outp(crtcaddr+4, 0xA0);
            temp = rr(crtcaddr, 0x33);
            outp(crtcaddr+1, 0xFF ^ temp);
            if ((rr(crtcaddr, 0x33) & 0xF0) == 0x00 &&
                  (rr(crtcaddr, 0x33) & 0x0F) == (temp ^ 0x0F))
               type = ET4000;
            outp(crtcaddr+1, temp);
            outp(crtcaddr+4, 0x29);
            }
         }
      else
         return (type);
      }
   temp = rr(crtcaddr, 0x1F) ^ rr(crtcaddr, 0x0C);
   if ((save = rr(0x3C4, 0x06)) > 1)
      return (type);
   outp(0x3C5, temp);
   if (rr(0x3C4, 0x06) != 1) {
      outp(0x3C5, save);
      return (type);
      }
   outp(0x3C5, (temp << 4)|(temp >> 4));
   if (rr(0x3C4, 0x06) != 0) {
      outp(0x3C5, save);
      return (type);
      }
   outp(0x3C5, save? temp: (temp << 4)|(temp >> 4));
   return (CIRRUS);
   }

unsigned int rr(port, reg)
   int port;
   int reg;
   {
   outp(port, reg);
   return (inp(port+1) & 0xFF);
   }

color256()
   {
   outp(0x3C6, 0x55);
   if ((inp(0x3C6) & 0xFF) != 0x55)
      return (0);
   outp(0x3C6, 0xAA);
   if ((inp(0x3C6) & 0xFF) != 0xAA)
      return (0);
   outp(0x3C6, 0x00);
   if ((inp(0x3C6) & 0xFF) != 0x00)
      return (0);
   outp(0x3C6, 0xFF);
   if ((inp(0x3C6) & 0xFF) != 0xFF)
      return (0);
   return (1);
   }

mem512()
   {
   unsigned int save, test;

   vidmode(128+46);
   save = peek(4, 0xA000);
   pokew(4, 0xA000, 0xA455);
   test = peek(4, 0xA000);
   if (test == 0xA455) {
      pokew(4, 0xA000, ~0xA455);
      test = peek(4, 0xA000);
      if (test == ~0xA455) {
         pokew(4, 0xA000, save);
         vidmode(3);
         return (1);
         }
      }
   vidmode(3);
   return (0);
   }
